@@ -133,9 +133,7 @@ ArgvMap &arg()
 struct timeval g_now;
 typedef vector<int> tcpserversocks_t;
 
-typedef MTasker<PacketID,string> MT_t;
-MT_t* MT;
-
+MT_t* MT; // the big MTasker
 
 void handleTCPClientWritable(int fd, boost::any& var);
 
@@ -272,7 +270,12 @@ public:
     if(i==d_socks.end()) {
       throw AhuException("Trying to return a socket not in the pool");
     }
-    g_fdm->removeReadFD(*i);
+    try {
+      g_fdm->removeReadFD(*i);
+    }
+    catch(FDMultiplexerException& e) {
+      // we sometimes return a socket that has not yet been assigned to g_fdm
+    }
     Utility::closesocket(*i);
     
     d_socks.erase(i++);
@@ -316,18 +319,22 @@ int asendto(const char *data, int len, int flags,
   
   int ret=connect(*fd, (struct sockaddr*)(&toaddr), toaddr.getSocklen());
   if(ret < 0) {
+    g_udpclientsocks.returnSocket(*fd);
     if(errno==ENETUNREACH) // Seth "My Interfaces Are Like A Yo Yo" Arnold special
       return -2;
     return ret;
   }
 
   g_fdm->addReadFD(*fd, handleUDPServerResponse, pident);
-  return send(*fd, data, len, 0);
+  ret=send(*fd, data, len, 0);
+  if(ret < 0)
+    g_udpclientsocks.returnSocket(*fd);
+  return ret;
 }
 
 // -1 is error, 0 is timeout, 1 is success
 int arecvfrom(char *data, int len, int flags, const ComboAddress& fromaddr, int *d_len, 
-	      uint16_t id, const string& domain, uint16_t qtype, int fd)
+	      uint16_t id, const string& domain, uint16_t qtype, int fd, unsigned int now)
 {
   static optional<unsigned int> nearMissLimit;
   if(!nearMissLimit) 
@@ -341,7 +348,7 @@ int arecvfrom(char *data, int len, int flags, const ComboAddress& fromaddr, int
   pident.remote=fromaddr;
 
   string packet;
-  int ret=MT->waitEvent(pident, &packet, 1);
+  int ret=MT->waitEvent(pident, &packet, 1, now);
 
   if(ret > 0) {
     if(packet.empty()) // means "error"
@@ -457,17 +464,22 @@ struct TCPConnection
   char data[65535];
   time_t startTime;
 
-  void closeAndCleanup()
+  static void closeAndCleanup(int fd, const ComboAddress& remote) 
   {
     Utility::closesocket(fd);
     if(!g_tcpClientCounts[remote]--) 
       g_tcpClientCounts.erase(remote);
     s_currentConnections--;
   }
+  void closeAndCleanup()
+  {
+    closeAndCleanup(fd, remote);
+  }
   static unsigned int s_currentConnections; //!< total number of current TCP connections
 };
 
 unsigned int TCPConnection::s_currentConnections; 
+void handleRunningTCPQuestion(int fd, boost::any& var);
 
 void startDoResolve(void *p)
 {
@@ -567,13 +579,17 @@ void startDoResolve(void *p)
 
       if(hadError) {
 	g_fdm->removeReadFD(dc->d_socket);
-  Utility::closesocket(dc->d_socket);
+	TCPConnection::closeAndCleanup(dc->d_socket, dc->d_remote);
       }
       else {
-	any_cast<TCPConnection>(&g_fdm->getReadParameter(dc->d_socket))->state=TCPConnection::BYTE0;
-	struct timeval now; 
-	Utility::gettimeofday(&now, 0); // needs to be updated
-	g_fdm->setReadTTD(dc->d_socket, now, g_tcpTimeout);
+	TCPConnection tc;
+	tc.fd=dc->d_socket;
+	tc.state=TCPConnection::BYTE0;
+	tc.remote=dc->d_remote;
+	Utility::gettimeofday(&g_now, 0); // needs to be updated
+	tc.startTime=g_now.tv_sec;
+	g_fdm->addReadFD(tc.fd, handleRunningTCPQuestion, tc);
+	g_fdm->setReadTTD(tc.fd, g_now, g_tcpTimeout);
       }
     }
 
@@ -630,13 +646,13 @@ void makeControlChannelSocket()
 void handleRunningTCPQuestion(int fd, boost::any& var)
 {
   TCPConnection* conn=any_cast<TCPConnection>(&var);
-  if(conn->state==TCPConnection::BYTE0) {
 
+  if(conn->state==TCPConnection::BYTE0) {
     int bytes=recv(conn->fd, conn->data, 2, 0);
     if(bytes==1)
       conn->state=TCPConnection::BYTE1;
     if(bytes==2) { 
-      conn->qlen=(conn->data[0]<<8)+conn->data[1];
+      conn->qlen=(((unsigned char)conn->data[0]) << 8)+ (unsigned char)conn->data[1];
       conn->bytesread=0;
       conn->state=TCPConnection::GETQUESTION;
     }
@@ -648,10 +664,10 @@ void handleRunningTCPQuestion(int fd, boost::any& var)
     }
   }
   else if(conn->state==TCPConnection::BYTE1) {
-    int bytes=recv(conn->fd,conn->data+1,1,0);
+    int bytes=recv(conn->fd, conn->data+1, 1, 0);
     if(bytes==1) {
       conn->state=TCPConnection::GETQUESTION;
-      conn->qlen=(conn->data[0]<<8)+conn->data[1];
+      conn->qlen=(((unsigned char)conn->data[0]) << 8)+ (unsigned char)conn->data[1];
       conn->bytesread=0;
     }
     if(!bytes || bytes < 0) {
@@ -664,7 +680,7 @@ void handleRunningTCPQuestion(int fd, boost::any& var)
     }
   }
   else if(conn->state==TCPConnection::GETQUESTION) {
-    int bytes=recv(conn->fd,conn->data + conn->bytesread,conn->qlen - conn->bytesread, 0);
+    int bytes=recv(conn->fd, conn->data + conn->bytesread, conn->qlen - conn->bytesread, 0);
     if(!bytes || bytes < 0) {
       L<<Logger::Error<<"TCP client "<< conn->remote.toString() <<" disconnected while reading question body"<<endl;
       TCPConnection tmp(*conn);
@@ -675,29 +691,34 @@ void handleRunningTCPQuestion(int fd, boost::any& var)
     }
     conn->bytesread+=bytes;
     if(conn->bytesread==conn->qlen) {
-      conn->state=TCPConnection::DONE;        // this makes us immune from timeouts, from now on *we* are responsible
+      TCPConnection tconn(*conn); 
+      g_fdm->removeReadFD(fd); // should no longer awake ourselves when there is data to read
+
       DNSComboWriter* dc=0;
       try {
-	dc=new DNSComboWriter(conn->data, conn->qlen, g_now);
+	dc=new DNSComboWriter(tconn.data, tconn.qlen, g_now);
       }
       catch(MOADNSException &mde) {
 	g_stats.clientParseError++; 
-	L<<Logger::Error<<"Unable to parse packet from TCP client "<< conn->remote.toString() <<endl;
-	TCPConnection tmp(*conn); 
-	g_fdm->removeReadFD(fd);
-	tmp.closeAndCleanup();
+	if(g_logCommonErrors)
+	  L<<Logger::Error<<"Unable to parse packet from TCP client "<< tconn.remote.toString() <<endl;
+	tconn.closeAndCleanup();
 	return;
       }
       
-      dc->setSocket(conn->fd);
+      dc->setSocket(tconn.fd);
       dc->d_tcp=true;
-      dc->setRemote(&conn->remote);
-      if(dc->d_mdp.d_header.qr)
+      dc->setRemote(&tconn.remote);
+      if(dc->d_mdp.d_header.qr) {
+	delete dc;
 	L<<Logger::Error<<"Ignoring answer on server socket!"<<endl;
+	tconn.closeAndCleanup();
+	return;
+      }
       else {
 	++g_stats.qcounter;
 	++g_stats.tcpqcounter;
-	MT->makeThread(startDoResolve, dc);
+	MT->makeThread(startDoResolve, dc); // deletes dc
 	return;
       }
     }
@@ -752,7 +773,6 @@ void handleNewUDPQuestion(int fd, boost::any& var)
   if((len=recvfrom(fd, data, sizeof(data), 0, (sockaddr *)&fromaddr, &addrlen)) >= 0) {
     g_stats.addRemote(fromaddr);
     if(g_allowFrom && !g_allowFrom->match(&fromaddr)) {
-      cout<<"mapped: "<<fromaddr.isMappedIPv4()<<endl;
       if(!g_quiet) 
 	L<<Logger::Error<<"["<<MT->getTid()<<"] dropping UDP query from "<<fromaddr.toString()<<", address not matched by allow-from"<<endl;
 
@@ -761,23 +781,24 @@ void handleNewUDPQuestion(int fd, boost::any& var)
     }
     try {
       DNSComboWriter* dc = new DNSComboWriter(data, len, g_now);
-      
       dc->setRemote(&fromaddr);
       
       if(dc->d_mdp.d_header.qr) {
 	if(g_logCommonErrors)
 	  L<<Logger::Error<<"Ignoring answer from "<<dc->getRemote()<<" on server socket!"<<endl;
+	delete dc;
       }
       else {
 	++g_stats.qcounter;
 	dc->setSocket(fd);
 	dc->d_tcp=false;
-	MT->makeThread(startDoResolve, (void*) dc);
+	MT->makeThread(startDoResolve, (void*) dc); // deletes dc
       }
     }
     catch(MOADNSException& mde) {
       g_stats.clientParseError++; 
-      L<<Logger::Error<<"Unable to parse packet from remote UDP client "<<fromaddr.toString() <<": "<<mde.what()<<endl;
+      if(g_logCommonErrors)
+	L<<Logger::Error<<"Unable to parse packet from remote UDP client "<<fromaddr.toString() <<": "<<mde.what()<<endl;
     }
   }
 }
@@ -794,14 +815,19 @@ void makeTCPServerSockets()
   if(locals.empty())
     throw AhuException("No local address specified");
   
-  ComboAddress sin;
   for(vector<string>::const_iterator i=locals.begin();i!=locals.end();++i) {
+    ServiceTuple st;
+    st.port=::arg().asNum("local-port");
+    parseService(*i, st);
+    
+    ComboAddress sin;
+
     memset((char *)&sin,0, sizeof(sin));
     sin.sin4.sin_family = AF_INET;
-    if(!IpToU32(*i, (uint32_t*)&sin.sin4.sin_addr.s_addr)) {
+    if(!IpToU32(st.host, (uint32_t*)&sin.sin4.sin_addr.s_addr)) {
       sin.sin6.sin6_family = AF_INET6;
-      if(Utility::inet_pton(AF_INET6, i->c_str(), &sin.sin6.sin6_addr) <= 0)
-	throw AhuException("Unable to resolve local address '"+ *i +"'"); 
+      if(Utility::inet_pton(AF_INET6, st.host.c_str(), &sin.sin6.sin6_addr) <= 0)
+	throw AhuException("Unable to resolve local address for TCP server on '"+ st.host +"'"); 
     }
 
     fd=socket(sin.sin6.sin6_family, SOCK_STREAM, 0);
@@ -821,20 +847,21 @@ void makeTCPServerSockets()
     }
 #endif
 
-    sin.sin4.sin_port = htons(::arg().asNum("local-port")); 
+    sin.sin4.sin_port = htons(st.port);
     int socklen=sin.sin4.sin_family==AF_INET ? sizeof(sin.sin4) : sizeof(sin.sin6);
     if (::bind(fd, (struct sockaddr *)&sin, socklen )<0) 
-      throw AhuException("Binding TCP server socket for "+*i+": "+stringerror());
+      throw AhuException("Binding TCP server socket for "+ st.host +": "+stringerror());
     
     Utility::setNonBlocking(fd);
     setSendBuffer(fd, 65000);
     listen(fd, 128);
     deferredAdd.push_back(make_pair(fd, handleNewTCPQuestion));
-    //    g_fdm->addReadFD(fd, handleNewTCPQuestion);
+    g_tcpListenSockets.push_back(fd);
+
     if(sin.sin4.sin_family == AF_INET) 
-      L<<Logger::Error<<"Listening for TCP queries on "<< sin.toString() <<":"<<::arg().asNum("local-port")<<endl;
+      L<<Logger::Error<<"Listening for TCP queries on "<< sin.toString() <<":"<<st.port<<endl;
     else
-      L<<Logger::Error<<"Listening for TCP queries on ["<< sin.toString() <<"]:"<<::arg().asNum("local-port")<<endl;
+      L<<Logger::Error<<"Listening for TCP queries on ["<< sin.toString() <<"]:"<<st.port<<endl;
   }
 }
 
@@ -851,14 +878,18 @@ void makeUDPServerSockets()
   }
 
   for(vector<string>::const_iterator i=locals.begin();i!=locals.end();++i) {
+    ServiceTuple st;
+    st.port=::arg().asNum("local-port");
+    parseService(*i, st);
+
     ComboAddress sin;
 
     memset(&sin, 0, sizeof(sin));
     sin.sin4.sin_family = AF_INET;
-    if(!IpToU32(*i, (uint32_t*)&sin.sin4.sin_addr.s_addr)) {
+    if(!IpToU32(st.host.c_str() , (uint32_t*)&sin.sin4.sin_addr.s_addr)) {
       sin.sin6.sin6_family = AF_INET6;
-      if(Utility::inet_pton(AF_INET6, i->c_str(), &sin.sin6.sin6_addr) <= 0)
-	throw AhuException("Unable to resolve local address '"+ *i +"'"); 
+      if(Utility::inet_pton(AF_INET6, st.host.c_str(), &sin.sin6.sin6_addr) <= 0)
+	throw AhuException("Unable to resolve local address for UDP server on '"+ st.host +"'"); 
     }
     
     int fd=socket(sin.sin4.sin_family, SOCK_DGRAM,0);
@@ -867,20 +898,20 @@ void makeUDPServerSockets()
     }
 
     setReceiveBuffer(fd, 200000);
-    sin.sin4.sin_port = htons(::arg().asNum("local-port")); 
+    sin.sin4.sin_port = htons(st.port);
 
     int socklen=sin.sin4.sin_family==AF_INET ? sizeof(sin.sin4) : sizeof(sin.sin6);
     if (::bind(fd, (struct sockaddr *)&sin, socklen)<0) 
-      throw AhuException("Resolver binding to server socket on port "+::arg()["local-port"]+" for "+*i+": "+stringerror());
+      throw AhuException("Resolver binding to server socket on port "+ lexical_cast<string>(st.port) +" for "+ st.host+": "+stringerror());
     
     Utility::setNonBlocking(fd);
     //    g_fdm->addReadFD(fd, handleNewUDPQuestion);
     deferredAdd.push_back(make_pair(fd, handleNewUDPQuestion));
-    g_tcpListenSockets.push_back(fd);
+
     if(sin.sin4.sin_family == AF_INET) 
-      L<<Logger::Error<<"Listening for UDP queries on "<< sin.toString() <<":"<<::arg().asNum("local-port")<<endl;
+      L<<Logger::Error<<"Listening for UDP queries on "<< sin.toString() <<":"<<st.port<<endl;
     else
-      L<<Logger::Error<<"Listening for UDP queries on ["<< sin.toString() <<"]:"<<::arg().asNum("local-port")<<endl;
+      L<<Logger::Error<<"Listening for UDP queries on ["<< sin.toString() <<"]:"<<st.port<<endl;
   }
 }
 
@@ -926,7 +957,7 @@ void usr2Handler(int)
 
 void doStats(void)
 {
-  if(g_stats.qcounter) {
+  if(g_stats.qcounter && (RC.cacheHits + RC.cacheMisses) && SyncRes::s_queries && SyncRes::s_outqueries) {
     L<<Logger::Error<<"stats: "<<g_stats.qcounter<<" questions, "<<RC.size()<<" cache entries, "<<SyncRes::s_negcache.size()<<" negative entries, "
      <<(int)((RC.cacheHits*100.0)/(RC.cacheHits+RC.cacheMisses))<<"% cache hits"<<endl;
     L<<Logger::Error<<"stats: throttle map: "<<SyncRes::s_throttle.size()<<", ns speeds: "
@@ -999,15 +1030,15 @@ catch(AhuException& ae)
 string questionExpand(const char* packet, uint16_t len, uint16_t& type)
 {
   type=0;
-  const char* end=packet+len;
-  const char* pos=packet+12;
+  const unsigned char* end=(const unsigned char*)packet+len;
+  const unsigned char* pos=(const unsigned char*)packet+12;
   unsigned char labellen;
   string ret;
   ret.reserve(len-12);
   while((labellen=*pos++)) {
     if(pos+labellen > end)
       break;
-    ret.append(pos, labellen);
+    ret.append((const char*)pos, labellen);
     ret.append(1,'.');
     pos+=labellen;
   }
@@ -1075,7 +1106,7 @@ void handleTCPClientWritable(int fd, boost::any& var)
 {
   PacketID* pid=any_cast<PacketID>(&var);
   
-  int ret=send(fd, pid->outMSG.c_str(), pid->outMSG.size() - pid->outPos,0);
+  int ret=send(fd, pid->outMSG.c_str() + pid->outPos, pid->outMSG.size() - pid->outPos,0);
   if(ret > 0) {
     pid->outPos+=ret;
     if(pid->outPos==pid->outMSG.size()) {
@@ -1159,8 +1190,8 @@ void handleUDPServerResponse(int fd, boost::any& var)
     }
 
     if(!MT->sendEvent(pident, &packet)) {
-      // if(g_logCommonErrors)
-      //   L<<Logger::Warning<<"Discarding unexpected packet from "<<sockAddrToString((struct sockaddr_in*) &fromaddr, addrlen)<<endl;
+//      if(g_logCommonErrors)
+//        L<<Logger::Warning<<"Discarding unexpected packet from "<<fromaddr.toString()<<": "<<pident.type<<endl;
       g_stats.unexpectedCount++;
       
       for(MT_t::waiters_t::iterator mthread=MT->d_waiters.begin(); mthread!=MT->d_waiters.end(); ++mthread) {
@@ -1416,7 +1447,16 @@ void parseAuthAndForwards()
 int serviceMain(int argc, char*argv[])
 {
   L.setName("pdns_recursor");
-  L.setLoglevel((Logger::Urgency)(4));
+
+  L.setLoglevel((Logger::Urgency)(6)); // info and up
+
+  if(!::arg()["logging-facility"].empty()) {
+    boost::optional<int> val=logFacilityToLOG(::arg().asNum("logging-facility") );
+    if(val)
+      theL().setFacility(*val);
+    else
+      L<<Logger::Error<<"Unknown logging facility "<<::arg().asNum("logging-facility") <<endl;
+  }
 
   L<<Logger::Warning<<"PowerDNS recursor "<<VERSION<<" (C) 2001-2006 PowerDNS.COM BV ("<<__DATE__", "__TIME__;
 #ifdef __GNUC__
@@ -1489,7 +1529,7 @@ int serviceMain(int argc, char*argv[])
   }
 #endif
   
-  MT=new MTasker<PacketID,string>(100000);
+  MT=new MTasker<PacketID,string>(::arg().asNum("stack-size"));
   makeControlChannelSocket();        
   PacketID pident;
   primeHints();    
@@ -1539,24 +1579,22 @@ int serviceMain(int argc, char*argv[])
   bool listenOnTCP(true);
   
   for(;;) {
-    while(MT->schedule()); // housekeeping, let threads do their thing
+    while(MT->schedule(g_now.tv_sec)); // housekeeping, let threads do their thing
       
     if(!(counter%500)) {
       MT->makeThread(houseKeeping,0);
     }
 
-    if(!(counter%11)) {
+    if(!(counter%55)) {
       typedef vector<pair<int, boost::any> > expired_t;
       expired_t expired=g_fdm->getTimeouts(g_now);
 	
       for(expired_t::iterator i=expired.begin() ; i != expired.end(); ++i) {
 	TCPConnection conn=any_cast<TCPConnection>(i->second);
-	if(conn.state != TCPConnection::DONE) {
-	  if(g_logCommonErrors)
-	    L<<Logger::Warning<<"Timeout from remote TCP client "<< conn.remote.toString() <<endl;
-	  g_fdm->removeReadFD(i->first);
-	  conn.closeAndCleanup();
-	}
+	if(g_logCommonErrors)
+	  L<<Logger::Warning<<"Timeout from remote TCP client "<< conn.remote.toString() <<endl;
+	g_fdm->removeReadFD(i->first);
+	conn.closeAndCleanup();
       }
     }
       
@@ -1620,12 +1658,13 @@ int main(int argc, char **argv)
 
   try {
     Utility::srandom(time(0));
+    ::arg().set("stack-size","stack size per mthread")="200000";
     ::arg().set("soa-minimum-ttl","Don't change")="0";
     ::arg().set("soa-serial-offset","Don't change")="0";
     ::arg().set("no-shuffle","Don't change")="off";
     ::arg().set("aaaa-additional-processing","turn on to do AAAA additional processing (slow)")="off";
     ::arg().set("local-port","port to listen on")="53";
-    ::arg().set("local-address","IP addresses to listen on, separated by spaces or commas")="127.0.0.1";
+    ::arg().set("local-address","IP addresses to listen on, separated by spaces or commas. Also accepts ports.")="127.0.0.1";
     ::arg().set("trace","if we should output heaps of logging")="off";
     ::arg().set("daemon","Operate as a daemon")="yes";
     ::arg().set("log-common-errors","If we should log rather common errors")="yes";
@@ -1642,6 +1681,7 @@ int main(int argc, char **argv)
     ::arg().setSwitch( "logfile", "Filename of the log file" )= "recursor.log"; 
 #else
     ::arg().set("quiet","Suppress logging of questions and answers")="";
+    ::arg().set("logging-facility","Facility to log messages as. 0 corresponds to local0")="";
 #endif
     ::arg().set("config-dir","Location of configuration directory (recursor.conf)")=SYSCONFDIR;
     ::arg().set("socket-dir","Where the controlsocket will live")=LOCALSTATEDIR;
@@ -1655,7 +1695,7 @@ int main(int argc, char **argv)
     ::arg().set("max-negative-ttl", "maximum number of seconds to keep a negative cached entry in memory")="3600";
     ::arg().set("server-id", "Returned when queried for 'server.id' TXT, defaults to hostname")="";
     ::arg().set("remotes-ringbuffer-entries", "maximum number of packets to store statistics for")="0";
-    ::arg().set("version-string", "string reported on version.pdns or version.bind")="PowerDNS Recursor "VERSION" $Id: pdns_recursor.cc 838 2006-05-19 14:35:27Z ahu $";
+    ::arg().set("version-string", "string reported on version.pdns or version.bind")="PowerDNS Recursor "VERSION" $Id: pdns_recursor.cc 917 2006-11-07 12:23:51Z ahu $";
     ::arg().set("allow-from", "If set, only allow these comma separated netmasks to recurse")="127.0.0.0/8, 10.0.0.0/8, 192.168.0.0/16, 172.16.0.0/12, ::1/128, fe80::/10";
     ::arg().set("max-tcp-per-client", "If set, maximum number of TCP sessions per client (IP address)")="0";
     ::arg().set("fork", "If set, fork the daemon for possible double performance")="no";
@@ -1666,10 +1706,11 @@ int main(int argc, char **argv)
     ::arg().set("export-etc-hosts", "If we should serve up contents from /etc/hosts")="off";
     ::arg().set("serve-rfc1918", "If we should be authoritative for RFC 1918 private IP space")="";
     ::arg().set("auth-can-lower-ttl", "If we follow RFC 2181 to the letter, an authoritative server can lower the TTL of NS records")="off";
+    ::arg().setSwitch( "ignore-rd-bit", "Assume each packet requires recursion, for compatability" )= "off"; 
 
     ::arg().setCmd("help","Provide a helpful message");
     ::arg().setCmd("config","Output blank configuration");
-    L.toConsole(Logger::Warning);
+    L.toConsole(Logger::Info);
     ::arg().laxParse(argc,argv); // do a lax parse
 
     string configname=::arg()["config-dir"]+"/recursor.conf";
